package net.knarcraft.ffmpegconverter.converter;

import net.knarcraft.ffmpegconverter.streams.AudioStream;
import net.knarcraft.ffmpegconverter.streams.StreamObject;
import net.knarcraft.ffmpegconverter.streams.SubtitleStream;
import net.knarcraft.ffmpegconverter.streams.VideoStream;

import java.io.File;
import java.io.IOException;
import java.util.List;

public class AnimeConverter extends Converter {
    private String[] audioLang;
    private String[] subtitleLang;
    private boolean toStereo;
    private boolean preventSignsAndSongs;

    /**
     * Instantiates a new anime converter
     * @param ffprobePath <p>Path/command to ffprobe.</p>
     * @param ffmpegPath <p>Path/command to ffmpeg.</p>
     * @param audioLang <p>List of wanted audio languages in descending order.</p>
     * @param subtitleLang <p>List of wanted subtitle languages in descending order.</p>
     * @param toStereo <p>Convert video with several audio channels to stereo.</p>
     * @param preventSignsAndSongs <p>Prevent subtitles only converting signs and songs (not speech).</p>
     */
    public AnimeConverter(String ffprobePath, String ffmpegPath, String[] audioLang, String[] subtitleLang,
                          boolean toStereo, boolean preventSignsAndSongs) {
        this.ffprobePath = ffprobePath;
        this.ffmpegPath = ffmpegPath;
        this.audioLang = audioLang;
        this.subtitleLang = subtitleLang;
        this.toStereo = toStereo;
        this.preventSignsAndSongs = preventSignsAndSongs;
    }

    /**
     * Converts the given file
     * @param file <p>The file to convert.</p>
     * @throws IOException <p>If the file cannot be converted.</p>
     */
    public void convert(File file) throws IOException {
        processFile(file.getParentFile(), file);
    }

    /**
     * Reads streams from a file, and converts it to an mp4
     * @param folder <p>The folder of the file to process.</p>
     * @param file <p>The file to process.</p>
     * @throws IOException <p>If the BufferedReader fails.</p>
     */
    private void processFile(File folder, File file) throws IOException {
        List<StreamObject> streams = probeFile(ffprobePath, file);
        if (streams.isEmpty()) {
            throw new IllegalArgumentException("The file has no valid streams. Please make sure the file exists and" +
                    " is not corrupt.");
        }
        String newPath = fileCollisionPrevention(folder.getAbsolutePath() + File.separator +
                stripExtension(file) + ".mp4", "mp4");
        printl("Preparing to start process...");
        String[] command = builderCommand(ffmpegPath, file.getName(), streams, newPath);
        ProcessBuilder processBuilder = new ProcessBuilder(command);
        convertProcess(processBuilder, folder);
    }

    /**
     * Generates a command for a ProcessBuilder.
     * @param executable <p>The executable file for ffmpeg.</p>
     * @param fileName <p>The input file.</p>
     * @param streams <p>A list of ffprobe streams.</p>
     * @param outFile <p>The output file.</p>
     * @return <p>A list of commands</p>
     */
    private String[] builderCommand(String executable, String fileName, List<StreamObject> streams, String outFile) {
        List<String> command = ffmpegWebVideo(executable, fileName);

        if (this.DEBUG) {
            addDebug(command, 50, 120);
        }

        List<AudioStream> audioStreams = filterAudioStreams(filterStreamsByType(streams, "audio"), audioLang);
        List<VideoStream> videoStreams = filterStreamsByType(streams, "video");
        List<SubtitleStream> subtitleStreams = filterSubtitleStreams(filterStreamsByType(streams,
                "subtitle"), subtitleLang, preventSignsAndSongs);

        VideoStream videoStream = null;
        AudioStream audioStream = null;
        SubtitleStream subtitleStream = null;
        if (videoStreams.size() > 0) {
            videoStream = videoStreams.get(0);
        }
        if (videoStream == null) {
            throw new IllegalArgumentException("The file does not have any valid video streams.");
        }
        if (audioStreams.size() > 0) {
            audioStream = audioStreams.get(0);
        }
        if (subtitleStreams.size() > 0) {
            subtitleStream = subtitleStreams.get(0);
        }

        addAudioStreams(command, audioStream);
        addSubtitles(command, subtitleStream, videoStream, fileName);

        command.add(outFile);
        return command.toArray(new String[0]);
    }

    /**
     * Adds audio to a command
     * @param command <p>The command to add audio to.</p>
     * @param audioStream <p>The audio stream to be added.</p>
     */
    private void addAudioStreams(List<String> command, AudioStream audioStream) {
        if (audioStream != null) {
            command.add("-map");
            command.add("0:" + audioStream.getAbsoluteIndex());
            if (toStereo && audioStream.getChannels() > 2) {
                command.add("-af");
                command.add("pan=stereo|FL=FC+0.30*FL+0.30*BL|FR=FC+0.30*FR+0.30*BR");
            }
        }
    }

    /**
     * Adds subtitles and video mapping to a command
     * @param command <p>The list containing the rest of the command.</p>
     * @param subtitleStream <p>The subtitle stream to be used.</p>
     * @param videoStream <p>The video stream to be used.</p>
     * @param fileName <p>The name of the file which is converted.</p>
     */
    private void addSubtitles(List<String> command, SubtitleStream subtitleStream, VideoStream videoStream,
                              String fileName) {
        if (subtitleStream != null && subtitleStream.getIsImageSubtitle()) {
            command.add("-filter_complex");
            String filter = String.format("[0:v:%d][0:%d]overlay", videoStream.getAbsoluteIndex(),
                    subtitleStream.getAbsoluteIndex());
            command.add(filter);
        } else if (subtitleStream != null) {
            command.add("-map");
            command.add(String.format("0:%d", videoStream.getAbsoluteIndex()));
            command.add("-vf");
            String safeFileName = escapeSpecialCharactersInFileName(fileName);
            String subtitleCommand = String.format("subtitles='%s':si=%d", safeFileName,
                    subtitleStream.getRelativeIndex());
            command.add(subtitleCommand);
        } else {
            command.add("-map");
            command.add(String.format("0:%d", videoStream.getAbsoluteIndex()));
        }
    }

    @Override
    public String[] getValidFormats() {
        return VIDEO_FORMATS;
    }
}
